
#include <nds.h>

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <math.h>

#include <fat.h>
#include <keyboard.h>
#include <dswifi9.h>

#include "bvideo.h"
#include "../lib/os_debug.h"


#define LDBFPRINTF_MACRO
#ifdef LDBFPRINTF_MACRO
#define ldbfprintf(fp, ...) fprintf(fp, __VA_ARGS__)
#else
extern int ldbfprintf(FILE *fp, const char *fmt,...);
#endif

void freeze() {
  while(1);
}

void *__stack_chk_guard = (void*)0xdeadbeef;
void __stack_chk_fail(void) { ldbfprintf(stderr, "Stack fail\n"); while(1); }


FILE* dsBaseModDebug = NULL;
static void (*ndsHeartbeatCallback)() = NULL;

void irqAcknowledge(u32 irqBits);

// Note: this is close to 0x20000000 which libnds uses (<.1% error)
#define BUS_CLK_HZ 33513982
#define D_SECONDS_PER_TICK ((double)1.0 / (double)BUS_CLK_HZ)
#define D_SECONDS_PER_ROLLOVER ((double)65536.0*D_SECONDS_PER_TICK)

void ihTimer0(void);
// FIXME: This will overflow about every 97 days
// (2^32 rolls/over)*(65536 ticks/roll)/(33513982 ticks/sec) = 8.4x10^6 secs/over
//  8.4x10^6 secs/over / (86400 secs/day) = 97 days
static volatile u32 timer0Rollovers = 0;
static volatile u32 timer1Rollovers = 0;
static u32 timer1RolloversLimit = 0;

void ihTimer1(void);
void ihVBlank(void);

static int loopCount = 0;
static int vBlankCount = 0;
static int cbCount = 0;

static unsigned char keyBuf[32];
static int keyCount = 0;

static int dbfprintfCount = 0;
extern void (*dbfprintfCallback)();

void initTimers() {
  // Timer 0 is the free running counter for clock ticks since start
  TIMER0_CR = 0;
  TIMER1_CR = 0;

  TIMER0_DATA = 0;
  TIMER1_DATA = 0;

  TIMER0_CR = TIMER_IRQ_REQ | TIMER_ENABLE;
}

void dbfprintfCounter() {
  dbfprintfCount++;

  if (0 == (dbfprintfCount & 0x7F)) {
    printf("\x1b[s\x1b[0;0H%d\x1b[u", dbfprintfCount);
  }
} 

void dumpBytes(const u8* src, u32 l) {
  static u32 ii = 0;

  for(ii=0; ii<l; ++ii) {
    printf("%02x", *(src+ii));
  }
}

void dumpData(const u8* src) {
  printf("\x1b[7;0H\n");

  for(int ii=0; ii<8; ++ii) {
    const u8* dd = src+(ii*65536);
    printf("%8p ", dd);
    dumpBytes(dd, 8);
    printf("\n");
  }
  
}

void waitForKeypress() {
  return;

  static int ii = 0;

  while(1) {
    scanKeys();

    if (keysDown() & KEY_DOWN) {
      break;
    }
  }

  for(ii=0; ii<1*60;++ii) {
    swiWaitForVBlank();
  }

  iprintf("Exiting wait\n");
  return;
}

FILE* cFOpen(const char* path, const char* mode) {
  FILE* stream = fopen(path, mode);
  if (NULL == stream) {
    iprintf("File open failed: %s %x %d\n", path, mode, stream, errno);
  } else {
    iprintf("File open succeded: %s %s\n", path, mode);
  }
  return stream;
}

int cFClose(FILE* stream) {
  int rvStatus = fclose(stream);
  if (0 == rvStatus) {
    iprintf("File close succeded\n");
  } else {
    iprintf("File close failed: %d %d\n", rvStatus, errno);
  }
  return rvStatus;
}

int getTouchKey(unsigned char* buf, int len) {
  char cc = 0;

  while(cc == 0) {
    scanKeys();
    swiWaitForVBlank();

    cc = processKeyboard((char*)buf, (unsigned int)1, (unsigned int)1);
    //    printf("\x1b[4;0HC:B:%d %02x:%02x  XX\n", loopCount, cc, *buf);

    if (cc < 32 || cc > 127) {
      if (cc != '\n') {
	// Translate all non-printable characters to 0
	cc = 0;
      }
    }

    ++loopCount;

    // processKeyboard() gets upset if you call it again too quickly
    swiWaitForVBlank();
  }

  *buf = cc;
  //    *(buf+1) = 0;

  //  printf("\x1b[5;0HC:B:%d %02x:%02x  XX\n", loopCount, cc, *buf);

  return 1;
}

// *********************** DSWIFI HELPERS **********************

// for some odd reason I can't seem to find vcount in ndslib -sigh-
// (I mean, in a place it doesn't conflict with other things)
#define         VCOUNT          (*((u16 volatile *) 0x04000006))


// wifi timer function, to update internals of sgIP
void Timer_50ms(void) {
   Wifi_Timer(50);
}

// notification function to send fifo message to arm7
void arm9_synctoarm7() { // send fifo message
   REG_IPC_FIFO_TX=0x87654321;
}

// interrupt handler to receive fifo messages from arm7
void arm9_fifo() { // check incoming fifo messages
   u32 value = REG_IPC_FIFO_RX;
   if(value == 0x87654321) Wifi_Sync();
}

void wifiInit() {
  // send fifo message to initialize the arm7 wifi
  REG_IPC_FIFO_CR = IPC_FIFO_ENABLE | IPC_FIFO_SEND_CLEAR; // enable & clear FIFO
                
  u32 Wifi_pass= Wifi_Init(WIFIINIT_OPTION_USELED);
  REG_IPC_FIFO_TX=0x12345678;
  REG_IPC_FIFO_TX=Wifi_pass;
        
  *((volatile u16 *)0x0400010E) = 0; // disable timer3
                
  irqSet(IRQ_TIMER3, Timer_50ms); // setup timer IRQ
  irqEnable(IRQ_TIMER3);
  irqSet(IRQ_FIFO_NOT_EMPTY, arm9_fifo); // setup fifo IRQ
  irqEnable(IRQ_FIFO_NOT_EMPTY);
        
  REG_IPC_FIFO_CR = IPC_FIFO_ENABLE | IPC_FIFO_RECV_IRQ; // enable FIFO IRQ
        
  Wifi_SetSyncHandler(arm9_synctoarm7); // tell wifi lib to use our handler to notify arm7

  // set timer3
  *((volatile u16 *)0x0400010C) = -6553; // 6553.1 * 256 cycles = ~50ms;
  *((volatile u16 *)0x0400010E) = 0x00C2; // enable, irq, 1/256 clock
  
  while(Wifi_CheckInit()==0) { // wait for arm7 to be initted successfully
    while(VCOUNT>192); // wait for vblank
    while(VCOUNT<192);
  }
  // wifi init complete - wifi lib can now be used!
}

void wifiAPConnect() {
  bool flgConnected = false;
  const int MAX_DHCP_RETRIES = 10;

  printf("Connecting with WFC data\n");

  for(int ii=0; ii<MAX_DHCP_RETRIES && !flgConnected; ++ii) {
    // Simple WFC connect -- Just sets flags -- does not block
    Wifi_AutoConnect();

    int st = 0;
    int stLast = 0;
    while(1) {
      st = Wifi_AssocStatus();

      if (st == ASSOCSTATUS_ASSOCIATED) {
	printf("Connected!            \n");
        flgConnected = true;
        break;
      }

      if (st == ASSOCSTATUS_CANNOTCONNECT) {
	printf("Failed...Retrying\n");
        break;
      }

      if (st != stLast) {
	printf("Connect status: %d %d\n", st, ii);
      }

      stLast = st;
      swiWaitForVBlank();
    }
  }

  if (!flgConnected) {
    printf("Failed...Terminating\n");
    exit(-1);
  }
}

// int main(void) {
int nongambit_main(int argc, char *argv[]) {
  static int ii = 0;

#ifdef ___DEBUG
  dbfprintfCallback = dbfprintfCounter;
#endif

  powerON(POWER_ALL);

  initVideoConsole();
  printf("\x1b[0;0HVideo init\n");

  irqInit();

  // Enable interrupt handlers
  irqSet(IRQ_VBLANK, ihVBlank);
  irqSet(IRQ_TIMER0, ihTimer0);
  irqSet(IRQ_TIMER1, ihTimer1);
  irqEnable(IRQ_VBLANK | IRQ_TIMER0 | IRQ_TIMER1);

  printf("IRQ init\n");

  initTimers();

  printf("Timer init\n");

  initKeyboard();

  printf("Key init\n");

#ifdef USE_DSWIFI
  wifiInit();
#endif

  printf("Wifi init\n");

/*   for(ii=0; ii<5*60; ++ii) { */
/*     // wait a couple second for DMA tfrs */
/*     swiWaitForVBlank(); */
/*   } */

/*   printf("Key wait\n"); */

/*   // FIXME: Even waiting on one, single key does something to the system */
/*   // which enables it to work */
/*   while(keyCount < 1) { */
/*     int numKeys = getTouchKey(keyBuf, 1); */
/*     ++keyCount; */
/*     printf("\x1b[6;0HKey: %d %d %d %02x  \n", keyCount, numKeys, keyBuf[0], keyBuf[0]); */
/*   } */


  printf("\x1b[7;0HTime: %d %d\n", timer0Rollovers, TIMER0_DATA);

  const char* fatFile = "fat2:/gambit.txt";

  // One cache page is about 550 bytes
  bool flgFATInit = fatInitDefault();
  if (flgFATInit) {
    iprintf("FAT initialized: %d\n", flgFATInit);
  } else {
    iprintf("FAT failed: %d\n", flgFATInit);
  }

  if (flgFATInit) {
    iprintf("Attempting to open file: %s\n", fatFile);

    FILE* fhFATFile = cFOpen(fatFile, "w");
    dsBaseModDebug = fhFATFile;

    for(ii=0; ii<60; ++ii) {
      // Sit here to prove that we are running
      touchPosition touchXY=touchReadXY();
      ldbfprintf(dsBaseModDebug, "\x1b[10;0Touch x = %04X, %04X\n", touchXY.x, touchXY.px);
      ldbfprintf(dsBaseModDebug, "Touch y = %04X, %04X\n", touchXY.y, touchXY.py);
      
      swiWaitForVBlank();
    }

/*     ldbfprintf(dsBaseModDebug, "Flushing..." */
/* 	      "                                      " */
/* 	      "                                      " */
/* 	      "                                      " */
/* 	      "                                      " */
/* 	      "                                      " */
/* 	      "                                      " */
/* 	      "                                      " */
/* 	      "                                      " */
/* 	      "                                      \n"); */
  } else {
    iprintf("Skipping FAT system\n");

    for(ii=0; ii<60; ++ii) {
      // Sit here to prove that we are running
      touchPosition touchXY=touchReadXY();
      ldbfprintf(dsBaseModDebug, "\x1b[10;0Touch x = %04X, %04X\n", touchXY.x, touchXY.px);
      dbfprintf(dsBaseModDebug, "Touch y = %04X, %04X\n", touchXY.y, touchXY.py);
      
      swiWaitForVBlank();
    }
  }

#ifdef USE_DSWIFI
  wifiAPConnect();
#endif

  u8* pp = (u8*)0x02000000;

  dumpData(pp);

  printf("Data: %d %d %d\n", *pp, *((u16*)pp), *((u32*)pp));

  iprintf("Exiting nds_main\n");

  // Return to the main entry for Scheme
  return 0;
}

inline void irqAcknowledge(u32 irqBits) {
  REG_IF = irqBits;
}

void ihVBlank(void) {
  ++vBlankCount;

  // Flag the fact that we have handled the interrupt
  irqAcknowledge(IRQ_VBLANK);  
}

void ihTimer0(void) {
  irqAcknowledge(IRQ_TIMER0);

  //  iprintf("\x1b[0;0HTime: %d %d\n", timer0Rollovers, TIMER0_DATA);
  timer0Rollovers = timer0Rollovers + 1;
}

void ihTimer1(void) {
  irqAcknowledge(IRQ_TIMER1);

  timer1Rollovers++;

  if (timer1Rollovers >= timer1RolloversLimit) {
    timer1Rollovers = 0;
    cbCount++;

    if (ndsHeartbeatCallback != NULL) {
      //      printf("\x1b[3;0HCalling back: %p %d   \x1b[21;0H", ndsHeartbeatCallback, cbCount);
      //      printf("Calling back: %p %d   \x1b[21;0H", ndsHeartbeatCallback, cbCount);
      //printf("Calling back: %p %d   \n", ndsHeartbeatCallback, cbCount);
      ndsHeartbeatCallback();
      //      printf("\x1b[4;0HReturning back: %p %d   \x1b[21;0H", ndsHeartbeatCallback, cbCount);
      //      printf("Returning back: %p %d   \x1b[21;0H", ndsHeartbeatCallback, cbCount);
      //printf("Returning back: %p %d   \n", ndsHeartbeatCallback, cbCount);
    }
  }
}

/* FIXME: WARNING! This is a *minimum* of an s32 in Gambit.  That
   does not mean it is *required* to be an s32.  It could be an s64 */
void ndsGetTime(s32 *secs, s32 *nsecs) {
  //printf("ndsGetTime entered\n");

  // Note: timer0Rollovers is *volatile*
  int rollovers = timer0Rollovers;
  int countTicks = TIMER0_DATA;
  int rolloversA = timer0Rollovers;

  if (rollovers != rolloversA) {
    /* Get the count again as we overflowed */
    /* It should not overflow again, but, if it does,
       then we are sufficiently slow that it probably doesn't matter */
    countTicks = TIMER0_DATA;
    rollovers = rolloversA;
  }

  double totalTime = ((double)65536.0*rollovers+countTicks)*D_SECONDS_PER_TICK;
  double integerSeconds = 0;
  double decimalSeconds = modf(totalTime, &integerSeconds);

  s32 integerNanoseconds = (s32)(decimalSeconds * (double)1.0e9);
  if (integerNanoseconds < 0) {
    integerNanoseconds = 0;
  }

  if (integerNanoseconds >= 1000000000) {
    integerNanoseconds = 999999999;
  }

  *secs = (s32)integerSeconds;
  *nsecs = (s32)integerNanoseconds;

  //printf("ndsGetTime exited\n");
}

double ndsSetHeartbeatInterval(double seconds) {
  printf("ndsSetHeart entered\n");

  if (seconds < 0.0) {
    // Disable timer1 completely
    TIMER1_CR = 0;

    printf("ndsSetHeart exited A\n");
    return seconds;
  } else if (seconds == 0.0) {
    timer1Rollovers = 0;
    timer1RolloversLimit = 500*2;

    TIMER1_DATA = 0;
    TIMER1_CR = TIMER_IRQ_REQ | TIMER_ENABLE;

    printf("ndsSetHeart exited B\n");
    return D_SECONDS_PER_ROLLOVER*timer1RolloversLimit;
  } else {
    u32 rolloverCount = (u32)(seconds / D_SECONDS_PER_ROLLOVER);
    timer1Rollovers = 0;
    timer1RolloversLimit = rolloverCount;

    TIMER1_DATA = 0;
    TIMER1_CR = TIMER_IRQ_REQ | TIMER_ENABLE;
    
    printf("ndsSetHeart exited C\n");
    return rolloverCount*D_SECONDS_PER_ROLLOVER;
  }
}

void ndsSetHeartbeatHandler(void (*heartbeatHandler)()) {
  printf("ndsHeartbeat entered: %p\n", heartbeatHandler);
  ndsHeartbeatCallback = heartbeatHandler;
  printf("ndsHeartbeat exited: %p\n", ndsHeartbeatCallback);
}
